import importlib
from gettext import gettext as _
import pkgutil

TRANSLATORS = {}
for _importer, modname, _ispkg in pkgutil.iter_modules(__path__):
    if modname != 'basetrans':
        modclass = importlib.import_module('dialect.translators.' + modname).Translator
        TRANSLATORS[modclass.name] = modclass


def check_backend_availability(backend_name):
    if backend_name in TRANSLATORS:
        return True

    return False


def get_fallback_backend_name():
    if TRANSLATORS:
        return next(iter(TRANSLATORS))
    return None


LANGUAGES = {
    'af': _('Afrikaans'),
    'sq': _('Albanian'),
    'am': _('Amharic'),
    'ar': _('Arabic'),
    'hy': _('Armenian'),
    'az': _('Azerbaijani'),
    'eu': _('Basque'),
    'be': _('Belarusian'),
    'bn': _('Bengali'),
    'bs': _('Bosnian'),
    'bg': _('Bulgarian'),
    'ca': _('Catalan'),
    'ceb': _('Cebuano'),
    'ny': _('Chichewa'),
    'zh': _('Chinese'),
    'zh-CN': _('Chinese (Simplified)'),
    'zh-TW': _('Chinese (Traditional)'),
    'zh_HANT': _('Chinese (Traditional)'),
    'co': _('Corsican'),
    'hr': _('Croatian'),
    'cs': _('Czech'),
    'da': _('Danish'),
    'nl': _('Dutch'),
    'en': _('English'),
    'eo': _('Esperanto'),
    'et': _('Estonian'),
    'tl': _('Filipino'),
    'fi': _('Finnish'),
    'fr': _('French'),
    'fy': _('Frisian'),
    'gl': _('Galician'),
    'ka': _('Georgian'),
    'de': _('German'),
    'el': _('Greek'),
    'gu': _('Gujarati'),
    'ht': _('Haitian Creole'),
    'ha': _('Hausa'),
    'haw': _('Hawaiian'),
    'iw': _('Hebrew'),
    'he': _('Hebrew'),
    'hi': _('Hindi'),
    'hmn': _('Hmong'),
    'hu': _('Hungarian'),
    'is': _('Icelandic'),
    'ig': _('Igbo'),
    'id': _('Indonesian'),
    'ga': _('Irish'),
    'it': _('Italian'),
    'ja': _('Japanese'),
    'jw': _('Javanese'),
    'kn': _('Kannada'),
    'kk': _('Kazakh'),
    'km': _('Khmer'),
    'rw': _('Kinyarwanda'),
    'ko': _('Korean'),
    'ku': _('Kurdish (Kurmanji)'),
    'ky': _('Kyrgyz'),
    'lo': _('Lao'),
    'la': _('Latin'),
    'lv': _('Latvian'),
    'lt': _('Lithuanian'),
    'lb': _('Luxembourgish'),
    'mk': _('Macedonian'),
    'mg': _('Malagasy'),
    'ms': _('Malay'),
    'ml': _('Malayalam'),
    'mt': _('Maltese'),
    'mi': _('Maori'),
    'mr': _('Marathi'),
    'mn': _('Mongolian'),
    'my': _('Myanmar (Burmese)'),
    'ne': _('Nepali'),
    'no': _('Norwegian'),
    'or': _('Odia (Oriya)'),
    'ps': _('Pashto'),
    'fa': _('Persian'),
    'pl': _('Polish'),
    'pt': _('Portuguese'),
    'pa': _('Punjabi'),
    'ro': _('Romanian'),
    'ru': _('Russian'),
    'sm': _('Samoan'),
    'gd': _('Scots Gaelic'),
    'sr': _('Serbian'),
    'st': _('Sesotho'),
    'sn': _('Shona'),
    'sd': _('Sindhi'),
    'si': _('Sinhala'),
    'sk': _('Slovak'),
    'sl': _('Slovenian'),
    'so': _('Somali'),
    'es': _('Spanish'),
    'su': _('Sundanese'),
    'sw': _('Swahili'),
    'sv': _('Swedish'),
    'tg': _('Tajik'),
    'ta': _('Tamil'),
    'tt': _('Tatar'),
    'te': _('Telugu'),
    'th': _('Thai'),
    'tr': _('Turkish'),
    'tk': _('Turkmen'),
    'uk': _('Ukrainian'),
    'ur': _('Urdu'),
    'ug': _('Uyghur'),
    'uz': _('Uzbek'),
    'vi': _('Vietnamese'),
    'cy': _('Welsh'),
    'xh': _('Xhosa'),
    'yi': _('Yiddish'),
    'yo': _('Yoruba'),
    'zu': _('Zulu'),
}


def get_lang_name(code):
    return LANGUAGES.get(code, code)



# Copyright 2020-2022 Mufeed Ali
# Copyright 2020-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import os
import re
import threading
from gettext import gettext as _

from gi.repository import Adw, Gio, GObject, Gtk, Soup

from dialect.define import RES_PATH
from dialect.session import Session
from dialect.settings import Settings
from dialect.translators import TRANSLATORS
from dialect.tts import TTS


@Gtk.Template(resource_path=f'{RES_PATH}/preferences.ui')
class DialectPreferencesWindow(Adw.PreferencesWindow):
    __gtype_name__ = 'DialectPreferencesWindow'

    parent = NotImplemented

    # Get preferences widgets
    live_translation = Gtk.Template.Child()
    sp_translation = Gtk.Template.Child()
    translate_accel = Gtk.Template.Child()
    src_auto = Gtk.Template.Child()
    backend = Gtk.Template.Child()
    instance_entry = Gtk.Template.Child()
    instance_stack = Gtk.Template.Child()
    instance_reset = Gtk.Template.Child()
    instance_spinner = Gtk.Template.Child()
    api_key_entry = Gtk.Template.Child()
    api_key_stack = Gtk.Template.Child()
    api_key_reset = Gtk.Template.Child()
    api_key_spinner = Gtk.Template.Child()
    tts = Gtk.Template.Child()
    tts_row = Gtk.Template.Child()
    search_provider = Gtk.Template.Child()

    def __init__(self, parent, **kwargs):
        super().__init__(**kwargs)

        self.parent = parent

        self.setup()

    def setup(self):
        # Disable search, we have few preferences
        self.set_search_enabled(False)

        # Setup backends combo row
        self.backend_model = Gio.ListStore.new(BackendObject)
        backend_options = [
            BackendObject(translator.name, translator.prettyname) for translator in TRANSLATORS.values()
        ]
        selected_backend_index = 0
        for index, value in enumerate(backend_options):
            self.backend_model.insert(index, value)
            if value.name == Settings.get().active_translator:
                selected_backend_index = index
        self.backend.set_model(self.backend_model)

        # Bind preferences with GSettings
        Settings.get().bind('live-translation', self.live_translation, 'enable-expansion',
                            Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('sp-translation', self.sp_translation, 'active',
                            Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('translate-accel', self.translate_accel,
                            'selected', Gio.SettingsBindFlags.DEFAULT)
        Settings.get().bind('src-auto', self.src_auto, 'active',
                            Gio.SettingsBindFlags.DEFAULT)

        # Setup TTS
        self.tts_row.set_visible(len(TTS) >= 1)
        self.tts.set_active(Settings.get().active_tts != '')

        # Set translate accel sensitivity by live translation state
        self.translate_accel.set_sensitive(not self.live_translation.get_enable_expansion())
        self.live_translation.connect('notify::enable-expansion', self._toggle_accel_pref)

        # Switch backends
        self.backend.set_selected(selected_backend_index)
        self.backend.connect('notify::selected', self._switch_backends)
        self.parent.connect('notify::backend-loading', self._on_backend_loading)

        # Toggle TTS
        self.tts.connect('notify::active', self._toggle_tts)

        # Connect to settings changes
        Settings.get().connect('changed', self._on_settings_changed)

        # Instance
        self.instance_entry.connect('apply', self._on_instance_apply)
        self.instance_entry.connect('changed', self._on_instance_changed)
        self.instance_reset.connect('clicked', self._on_reset_instance)

        # API Key
        self.api_key_entry.connect('apply', self._on_api_key_apply)
        self.api_key_reset.connect('clicked', self._on_reset_api_key)

        self.__check_instance_or_api_key_support()

        # Search Provider
        if os.getenv('XDG_CURRENT_DESKTOP') != 'GNOME':
            self.search_provider.hide()

    def _unbind_settings(self,  *args, **kwargs):
        Settings.get().unbind(self.live_translation, 'active')
        Settings.get().unbind(self.src_auto, 'active')

    def _on_settings_changed(self, _settings, key):
        backend = Settings.get().active_translator
        if key in ('translator-instance-url', 'translator-api-key'):
            # Update backend
            if key == 'translator-instance-url' and TRANSLATORS[backend].supported_features['change-instance']:
                Settings.get().reset_src_langs()
                Settings.get().reset_dest_langs()
            self.parent.reload_backends()

    def _toggle_accel_pref(self, row, _active):
        self.translate_accel.set_sensitive(not row.get_enable_expansion())

    def _toggle_tts(self, switch, _active):
        value = ''
        if switch.get_active() and len(TTS) >= 1:
            tts = list(TTS.keys())
            value = str(tts[0])

        self.parent.src_voice_btn.set_sensitive(False)
        self.parent.src_voice_btn.set_visible(switch.get_active())
        self.parent.dest_voice_btn.set_sensitive(False)
        self.parent.dest_voice_btn.set_visible(switch.get_active())

        Settings.get().active_tts = value

        if switch.get_active():
            threading.Thread(
                target=self.parent.load_lang_speech,
                daemon=True
            ).start()

    def _switch_backends(self, row, _value):
        self.parent.save_settings()
        backend = self.backend_model[row.get_selected()].name
        Settings.get().active_translator = backend
        self.__check_instance_or_api_key_support()
        self.parent.reload_backends()

    def _on_backend_loading(self, window, _value):
        self.backend.set_sensitive(not window.get_property('backend-loading'))
        self.instance_entry.set_sensitive(not window.get_property('backend-loading'))
        self.api_key_entry.set_sensitive(not window.get_property('backend-loading'))

        if not window.get_property('backend-loading'):
            self.instance_stack.set_visible_child_name('reset')
            self.instance_spinner.stop()
            self.api_key_stack.set_visible_child_name('reset')
            self.api_key_spinner.stop()

            # Show or hide api key entry
            if window.translator:
                if window.translator.supported_features['api-key-supported']:
                    self.api_key_entry.set_visible(True)
                    self.api_key_entry.set_text(Settings.get().api_key)
                else:
                    self.api_key_entry.set_visible(False)

    def _on_instance_apply(self, _row):
        def on_validation_response(session, result):
            valid = False
            backend = Settings.get().active_translator
            try:
                data = Session.get_response(session, result)
                valid = TRANSLATORS[backend].validate_instance(data)
            except Exception as exc:
                logging.error(exc)

            if valid:
                Settings.get().instance_url = self.new_instance_url
                self.instance_entry.remove_css_class('error')
                self.instance_entry.set_text(Settings.get().instance_url)
            else:
                self.instance_entry.add_css_class('error')
                error_text = _('Not a valid {backend} instance')
                error_text = error_text.format(backend=TRANSLATORS[backend].prettyname)
                toast = Adw.Toast.new(error_text)
                self.add_toast(toast)
                self.api_key_entry.set_visible(False)
                self.api_key_entry.set_text('')

            self.backend.set_sensitive(True)
            self.instance_entry.set_sensitive(True)
            self.api_key_entry.set_sensitive(True)
            self.instance_stack.set_visible_child_name('reset')
            self.instance_spinner.stop()

        old_value = Settings.get().instance_url
        new_value = self.instance_entry.get_text()

        url = re.compile(r'https?://(www\.)?')
        self.new_instance_url = url.sub('', new_value).strip().strip('/')

        # Validate
        if self.new_instance_url != old_value:
            # Progress feedback
            self.backend.set_sensitive(False)
            self.instance_entry.set_sensitive(False)
            self.api_key_entry.set_sensitive(False)
            self.instance_stack.set_visible_child_name('spinner')
            self.instance_spinner.start()

            backend = Settings.get().active_translator
            validation_url = TRANSLATORS[backend].format_url(
                self.new_instance_url,
                TRANSLATORS[backend].validation_path
            )
            validation_message = Soup.Message.new('GET', validation_url)

            Session.get().send_and_read_async(validation_message, 0, None, on_validation_response)
        else:
            self.instance_entry.remove_css_class('error')

    def _on_instance_changed(self, _entry):
        if self.instance_entry.props.text == Settings.get().instance_url:
            self.instance_entry.props.show_apply_button = False
        elif not self.instance_entry.props.show_apply_button:
            self.instance_entry.props.show_apply_button = True

    def _on_reset_instance(self, _button):
        backend = Settings.get().active_translator
        if Settings.get().instance_url != TRANSLATORS[backend].instance_url:
            Settings.get().reset_instance_url()
            Settings.get().reset_api_key()
            self.instance_stack.set_visible_child_name('spinner')
            self.instance_spinner.start()
        self.instance_entry.remove_css_class('error')
        self.instance_entry.set_text(Settings.get().instance_url)

    def _on_api_key_apply(self, _row):
        def on_response(session, result):
            valid = False
            try:
                data = Session.get_response(session, result)
                self.parent.translator.get_translation(data)
                valid = True
            except Exception as exc:
                logging.warning(exc)

            if valid:
                Settings.get().api_key = self.new_api_key
                self.api_key_entry.remove_css_class('error')
                self.api_key_entry.set_text(Settings.get().instance_url)
            else:
                self.api_key_entry.add_css_class('error')
                error_text = _('Not a valid {backend} API key')
                error_text = error_text.format(backend=TRANSLATORS[backend].prettyname)
                toast = Adw.Toast.new(error_text)
                self.add_toast(toast)
                self.api_key_entry.grab_focus()

            self.backend.set_sensitive(True)
            self.instance_entry.set_sensitive(True)
            self.api_key_entry.set_sensitive(True)
            self.api_key_stack.set_visible_child_name('reset')
            self.api_key_spinner.stop()

        old_value = Settings.get().api_key
        self.new_api_key = self.api_key_entry.get_text()

        if self.new_api_key != old_value:
            # Progress feedback
            self.backend.set_sensitive(False)
            self.instance_entry.set_sensitive(False)
            self.api_key_entry.set_sensitive(False)
            self.api_key_stack.set_visible_child_name('spinner')
            self.api_key_spinner.start()

            backend = Settings.get().active_translator
            validation_url = TRANSLATORS[backend].format_url(
                Settings.get().instance_url,
                TRANSLATORS[backend].api_test_path
            )
            (data, headers) = TRANSLATORS[backend].format_api_key_test(self.new_api_key)
            message = Session.create_message('POST', validation_url, data, headers)

            Session.get().send_and_read_async(message, 0, None, on_response)
        else:
            self.api_key_entry.remove_css_class('error')

    def _on_reset_api_key(self, _button):
        backend = Settings.get().active_translator
        if Settings.get().api_key != TRANSLATORS[backend].api_key:
            Settings.get().reset_api_key()
            self.api_key_stack.set_visible_child_name('spinner')
            self.api_key_spinner.start()
        self.api_key_entry.remove_css_class('error')
        self.api_key_entry.set_text(Settings.get().api_key)

    def __check_instance_or_api_key_support(self):
        backend = Settings.get().active_translator
        if TRANSLATORS[backend].supported_features['change-instance']:
            self.instance_entry.set_visible(True)
            self.instance_entry.set_text(Settings.get().instance_url)
        else:
            self.instance_entry.set_visible(False)

        if TRANSLATORS[backend].supported_features['api-key-supported']:
            self.api_key_entry.set_visible(True)
            self.api_key_entry.set_text(Settings.get().api_key)
        else:
            self.api_key_entry.set_visible(False)


class BackendObject(GObject.Object):
    __gtype_name__ = 'BackendObject'

    name = GObject.Property(type=str)
    prettyname = GObject.Property(type=str)

    def __init__(self, name, prettyname):
        super().__init__()

        self.set_property('name', name)
        self.set_property('prettyname', prettyname)



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import json
import logging

from gi.repository import GLib, Soup


class Session(Soup.Session):
    """
    Dialect soup session handler
    """

    instance = None

    def __init__(self):
        Soup.Session.__init__(self)

    @staticmethod
    def new():
        """Create a new instance of Session."""
        s_session = Soup.Session()
        s_session.__class__ = Session
        return s_session

    @staticmethod
    def get():
        """Return an active instance of Session."""
        if Session.instance is None:
            Session.instance = Session.new()
        return Session.instance

    @staticmethod
    def get_response(session, result, fail_if_empty=True, raw=False):
        try:
            response = session.send_and_read_finish(result)

            data = Session.read_response(response, fail_if_empty, raw)

            return data
        except GLib.GError as exc:
            raise ResponseError(exc.message) from exc

    @staticmethod
    def read_response(response, fail_if_empty=True, raw=False):
        if raw:
            return response.get_data()

        data = {}
        try:
            data = json.loads(
                response.get_data()
            ) if response else {}
        except Exception as exc:
            logging.warning(exc)

        if not data and fail_if_empty:
            raise ResponseEmpty()

        return data

    @staticmethod
    def encode_data(data):
        data_glib_bytes = None
        try:
            data_bytes = json.dumps(data).encode('utf-8')
            data_glib_bytes = GLib.Bytes.new(data_bytes)
        except Exception as exc:
            logging.warning(exc)
        return data_glib_bytes

    @staticmethod
    def create_message(method, url, data={}, headers={}, form=False):
        if form and data:
            form_data = Soup.form_encode_hash(data)
            message = Soup.Message.new_from_encoded_form(method, url, form_data)
        else:
            message = Soup.Message.new(method, url)
        if data and not form:
            data = Session.encode_data(data)
            message.set_request_body_from_bytes('application/json', data)
        if headers:
            for name, value in headers.items():
                message.get_request_headers().append(name, value)
        if 'User-Agent' not in headers:
            message.get_request_headers().append('User-Agent', 'Dialect App')
        return message

    def multiple(self, messages, callback=None):
        """Keep track of multiple async operations."""
        def on_task_response(session, result, message_callback):
            messages.pop()
            message_callback(session, result)

            # If all tasks are done, run main callback
            if callback is not None and len(messages) == 0:
                callback()

        for msg in messages:
            # msg[0]: Soup.Message
            # msg[1]: message callback
            self.send_and_read_async(msg[0], 0, None, on_task_response, msg[1])


class ResponseError(Exception):
    """Exception raised when response fails."""

    def __init__(self, cause, message='Response has failed'):
        self.cause = cause
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}: {self.cause}'


class ResponseEmpty(Exception):
    """Exception raised when response is empty."""

    def __init__(self, message='Response is empty'):
        self.message = message
        super().__init__(self.message)



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

class TextToSpeechBase:
    name = ''
    prettyname = ''
    languages = ['en']

    def download_voice(self, text, language, file):
        pass


class TextToSpeechError(Exception):
    """Exception raised when tts fails."""

    def __init__(self, cause, message='Text to Speech has failed'):
        self.cause = cause
        self.message = message
        super().__init__(self.message)



# Copyright 2020 gi-lom
# Copyright 2020-2022 Mufeed Ali
# Copyright 2020-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

# Initial setup
import sys
from gettext import gettext as _

import gi
gi.require_version('Gdk', '4.0')
gi.require_version('Gtk', '4.0')
gi.require_version('Gst', '1.0')
gi.require_version('Adw', '1')
gi.require_version('Soup', '3.0')

from gi.repository import Adw, Gio, GLib, Gst, Gtk

from dialect.define import APP_ID, RES_PATH, VERSION
from dialect.preferences import DialectPreferencesWindow
from dialect.settings import Settings
from dialect.window import DialectWindow


class Dialect(Adw.Application):
    def __init__(self):
        Adw.Application.__init__(
            self,
            application_id=APP_ID,
            flags=Gio.ApplicationFlags.HANDLES_COMMAND_LINE
        )
        self.set_resource_base_path(RES_PATH)

        # App window
        self.window = None
        self.launch_text = ''
        self.launch_langs = {}

        # Add command line options
        self.add_main_option('text', b't', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Text to translate', None)
        self.add_main_option('src', b's', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Source lang code', None)
        self.add_main_option('dest', b'd', GLib.OptionFlags.NONE,
                             GLib.OptionArg.STRING, 'Destination lang code', None)

        self.setup_actions()

    def do_activate(self):
        self.window = self.props.active_window

        if not self.window:
            width, height = Settings.get().window_size
            self.window = DialectWindow(
                application=self,
                # Translators: Do not translate the app name!
                title=_('Dialect'),
                default_height=height,
                default_width=width,
                text=self.launch_text,
                langs=self.launch_langs
            )

        self.window.present()

    def do_command_line(self, command_line):
        options = command_line.get_options_dict()
        options = options.end().unpack()
        text = ''
        langs = {
            'src': None,
            'dest': None
        }

        if 'text' in options:
            text = options['text']
        if 'src' in options:
            langs['src'] = options['src']
        if 'dest' in options:
            langs['dest'] = options['dest']

        if self.window is not None:
            self.window.translate(text, langs['src'], langs['dest'])
        else:
            self.launch_text = text
            self.launch_langs = langs

        self.activate()
        return 0

    def do_startup(self):
        Adw.Application.do_startup(self)

        Gst.init(None)  # Init Gst

    def setup_actions(self):
        """ Setup menu actions """

        pronunciation = Gio.SimpleAction.new_stateful(
            'pronunciation', None, Settings.get().show_pronunciation_value
        )
        pronunciation.connect('change-state', self._on_pronunciation)
        self.add_action(pronunciation)

        preferences = Gio.SimpleAction.new('preferences', None)
        preferences.connect('activate', self._on_preferences)
        self.add_action(preferences)

        about = Gio.SimpleAction.new('about', None)
        about.connect('activate', self._on_about)
        self.add_action(about)

        quit_action = Gio.SimpleAction.new('quit', None)
        quit_action.connect('activate', self._on_quit)
        self.add_action(quit_action)

        self.set_accels_for_action('app.pronunciation', ['<Primary>P'])
        self.set_accels_for_action('app.preferences', ['<Primary>comma'])
        self.set_accels_for_action('app.quit', ['<Primary>Q'])

        self.set_accels_for_action('win.back', ['<Alt>Left'])
        self.set_accels_for_action('win.forward', ['<Alt>Right'])
        self.set_accels_for_action('win.switch', ['<Primary>S'])
        self.set_accels_for_action('win.clear', ['<Primary>D'])
        self.set_accels_for_action('win.paste', ['<Primary><Shift>V'])
        self.set_accels_for_action('win.copy', ['<Primary><Shift>C'])
        self.set_accels_for_action('win.listen-dest', ['<Primary>L'])
        self.set_accels_for_action('win.listen-src', ['<Primary><Shift>L'])
        self.set_accels_for_action('win.show-help-overlay', ['<Primary>question'])

    def _on_pronunciation(self, action, value):
        """ Update show pronunciation setting """
        action.set_state(value)
        Settings.get().show_pronunciation = value

        # Update UI
        if self.window.trans_src_pron is not None:
            self.window.src_pron_revealer.set_reveal_child(value)
        if self.window.trans_dest_pron is not None:
            self.window.dest_pron_revealer.set_reveal_child(value)

    def _on_preferences(self, _action, _param):
        """ Show preferences window """
        window = DialectPreferencesWindow(self.window)
        window.set_transient_for(self.window)
        window.present()

    def _on_about(self, _action, _param):
        """ Show about dialog """
        builder = Gtk.Builder.new_from_resource(f'{RES_PATH}/about.ui')
        about = builder.get_object('about')

        about.set_application_icon(APP_ID)
        about.set_version(VERSION)
        about.set_developers(
            [
                "Mufeed Ali",
                "Rafael Mardojai CM http://rafaelmardojai.com"
            ]
        )

        about.set_transient_for(self.window)
        about.present()

    def _on_quit(self, _action, _param):
        self.quit()


def main():
    # Run the Application
    app = Dialect()
    return app.run(sys.argv)



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from gtts import gTTS, lang

from dialect.tts.basetts import TextToSpeechBase, TextToSpeechError


class TextToSpeech(TextToSpeechBase):
    name = 'google'
    prettyname = 'Google Text-to-Speech'
    languages = []

    def __init__(self, **kwargs):
        self.languages = list(lang.tts_langs().keys())

    def download_voice(self, text, language, file):
        try:
            tts = gTTS(text, lang=language, lang_check=False)
            tts.write_to_fp(file)
            file.seek(0)

        except Exception as exc:
            raise TextToSpeechError(exc) from exc



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

class TranslatorBase:
    name = ''
    prettyname = ''
    history = []
    languages = ['en']
    chars_limit = 0
    supported_features = {
        'detection': False,
        'mistakes': False,
        'pronunciation': False,
        'change-instance': False,
        'suggestions': False,
        'api-key-supported': False,
        'api-key-required': False,
    }
    instance_url = ''
    api_key = ''
    src_langs = ['en', 'fr', 'es', 'de']
    dest_langs = ['fr', 'es', 'de', 'en']

    validation_path = ''
    api_test_path = ''

    @staticmethod
    def format_url(url, path, http=False):
        protocol = 'https://'
        if url.startswith('localhost:') or http:
            protocol = 'http://'

        return protocol + url + path

    @staticmethod
    def validate_instance(data):
        pass

    def format_suggestion(self, text, src, dest, suggestion):
        pass

    def get_suggestion(self, data):
        pass

    def format_translation(self, text, src, dest):
        pass

    def get_translation(self, data):
        pass


class TranslatorError(Exception):
    """Base Exception for Translator related errors."""

    def __init__(self, cause, message='Translator Error'):
        self.cause = cause
        self.message = message
        super().__init__(self.message)

    def __str__(self):
        return f'{self.message}: {self.cause}'


class ApiKeyRequired(TranslatorError):
    """Exception raised when API key is required."""

    def __init__(self, cause, message='API Key Required'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class InvalidApiKey(TranslatorError):
    """Exception raised when an invalid API key is found."""

    def __init__(self, cause, message='Invalid API Key'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class InvalidLangCode(TranslatorError):
    """Exception raised when an invalid lang code is sent."""

    def __init__(self, cause, message='Invalid Lang Code'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class BatchSizeExceeded(TranslatorError):
    """Exception raised when the batch size limit has been exceeded."""

    def __init__(self, cause, message='Batch Size Exceeded'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class CharactersLimitExceeded(TranslatorError):
    """Exception raised when the char limit has been exceeded."""

    def __init__(self, cause, message='Characters Limit Exceeded'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class ServiceLimitReached(TranslatorError):
    """Exception raised when the service limit has been reached."""

    def __init__(self, cause, message='Service Limit Reached'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class TranslationError(TranslatorError):
    """Exception raised when translation fails."""

    def __init__(self, cause, message='Translation has failed'):
        self.cause = cause
        self.message = message
        super().__init__(self.cause, self.message)


class Translation:
    text = None
    extra_data = {
        'possible-mistakes': None,
        'src-pronunciation': None,
        'dest-pronunciation': None,
    }

    def __init__(self, text, extra_data):
        self.text = text
        self.extra_data = extra_data



# Copyright 2022 Mufeed Ali
# Copyright 2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import html
import json
import random
import re
from typing import List

from dialect.translators.basetrans import TranslatorBase, Translation, TranslationError, TranslatorError

RPC_ID = 'MkEWBc'

# Predefined URLs used to make google translate requests.
TRANSLATE_RPC = '{host}/_/TranslateWebserverUi/data/batchexecute'

DEFAULT_SERVICE_URLS = ('translate.google.ac', 'translate.google.ad', 'translate.google.ae',
                        'translate.google.al', 'translate.google.am', 'translate.google.as',
                        'translate.google.at', 'translate.google.az', 'translate.google.ba',
                        'translate.google.be', 'translate.google.bf', 'translate.google.bg',
                        'translate.google.bi', 'translate.google.bj', 'translate.google.bs',
                        'translate.google.bt', 'translate.google.by', 'translate.google.ca',
                        'translate.google.cat', 'translate.google.cc', 'translate.google.cd',
                        'translate.google.cf', 'translate.google.cg', 'translate.google.ch',
                        'translate.google.ci', 'translate.google.cl', 'translate.google.cm',
                        'translate.google.cn', 'translate.google.co.ao', 'translate.google.co.bw',
                        'translate.google.co.ck', 'translate.google.co.cr', 'translate.google.co.id',
                        'translate.google.co.il', 'translate.google.co.in', 'translate.google.co.jp',
                        'translate.google.co.ke', 'translate.google.co.kr', 'translate.google.co.ls',
                        'translate.google.co.ma', 'translate.google.co.mz', 'translate.google.co.nz',
                        'translate.google.co.th', 'translate.google.co.tz', 'translate.google.co.ug',
                        'translate.google.co.uk', 'translate.google.co.uz', 'translate.google.co.ve',
                        'translate.google.co.vi', 'translate.google.co.za', 'translate.google.co.zm',
                        'translate.google.co.zw', 'translate.google.com.af', 'translate.google.com.ag',
                        'translate.google.com.ai', 'translate.google.com.ar', 'translate.google.com.au',
                        'translate.google.com.bd', 'translate.google.com.bh', 'translate.google.com.bn',
                        'translate.google.com.bo', 'translate.google.com.br', 'translate.google.com.bz',
                        'translate.google.com.co', 'translate.google.com.cu', 'translate.google.com.cy',
                        'translate.google.com.do', 'translate.google.com.ec', 'translate.google.com.eg',
                        'translate.google.com.et', 'translate.google.com.fj', 'translate.google.com.gh',
                        'translate.google.com.gi', 'translate.google.com.gt', 'translate.google.com.hk',
                        'translate.google.com.jm', 'translate.google.com.kh', 'translate.google.com.kw',
                        'translate.google.com.lb', 'translate.google.com.ly', 'translate.google.com.mm',
                        'translate.google.com.mt', 'translate.google.com.mx', 'translate.google.com.my',
                        'translate.google.com.na', 'translate.google.com.ng', 'translate.google.com.ni',
                        'translate.google.com.np', 'translate.google.com.om', 'translate.google.com.pa',
                        'translate.google.com.pe', 'translate.google.com.pg', 'translate.google.com.ph',
                        'translate.google.com.pk', 'translate.google.com.pr', 'translate.google.com.py',
                        'translate.google.com.qa', 'translate.google.com.sa', 'translate.google.com.sb',
                        'translate.google.com.sg', 'translate.google.com.sl', 'translate.google.com.sv',
                        'translate.google.com.tj', 'translate.google.com.tr', 'translate.google.com.tw',
                        'translate.google.com.ua', 'translate.google.com.uy', 'translate.google.com.vc',
                        'translate.google.com.vn', 'translate.google.com', 'translate.google.cv',
                        'translate.google.cz', 'translate.google.de', 'translate.google.dj',
                        'translate.google.dk', 'translate.google.dm', 'translate.google.dz',
                        'translate.google.ee', 'translate.google.es', 'translate.google.fi',
                        'translate.google.fm', 'translate.google.fr', 'translate.google.ga',
                        'translate.google.ge', 'translate.google.gg', 'translate.google.gl',
                        'translate.google.gm', 'translate.google.gp', 'translate.google.gr',
                        'translate.google.gy', 'translate.google.hn', 'translate.google.hr',
                        'translate.google.ht', 'translate.google.hu', 'translate.google.ie',
                        'translate.google.im', 'translate.google.iq', 'translate.google.is',
                        'translate.google.it', 'translate.google.je', 'translate.google.jo',
                        'translate.google.kg', 'translate.google.ki', 'translate.google.kz',
                        'translate.google.la', 'translate.google.li', 'translate.google.lk',
                        'translate.google.lt', 'translate.google.lu', 'translate.google.lv',
                        'translate.google.md', 'translate.google.me', 'translate.google.mg',
                        'translate.google.mk', 'translate.google.ml', 'translate.google.mn',
                        'translate.google.ms', 'translate.google.mu', 'translate.google.mv',
                        'translate.google.mw', 'translate.google.ne', 'translate.google.nl',
                        'translate.google.no', 'translate.google.nr', 'translate.google.nu',
                        'translate.google.pl', 'translate.google.pn', 'translate.google.ps',
                        'translate.google.pt', 'translate.google.ro', 'translate.google.rs',
                        'translate.google.ru', 'translate.google.rw', 'translate.google.sc',
                        'translate.google.se', 'translate.google.sh', 'translate.google.si',
                        'translate.google.sk', 'translate.google.sm', 'translate.google.sn',
                        'translate.google.so', 'translate.google.sr', 'translate.google.st',
                        'translate.google.td', 'translate.google.tg', 'translate.google.tk',
                        'translate.google.tl', 'translate.google.tm', 'translate.google.tn',
                        'translate.google.to', 'translate.google.tt', 'translate.google.us',
                        'translate.google.vg', 'translate.google.vu', 'translate.google.ws')


class Translator(TranslatorBase):
    name = 'google'
    prettyname = 'Google Translate'
    client = None
    history = []
    languages = [
        'af', 'sq', 'am', 'ar', 'hy', 'az', 'eu', 'be', 'bn', 'bs', 'bg', 'ca',
        'ceb', 'ny', 'zh-CN', 'zh-TW', 'co', 'hr', 'cs', 'da', 'nl', 'en', 'eo',
        'et', 'tl', 'fi', 'fr', 'fy', 'gl', 'ka', 'de', 'el', 'gu', 'ht', 'ha',
        'haw', 'iw', 'hi', 'hmn', 'hu', 'is', 'ig', 'id', 'ga', 'it', 'ja', 'jw',
        'kn', 'kk', 'km', 'rw', 'ko', 'ku', 'ky', 'lo', 'la', 'lv', 'lt', 'lb',
        'mk', 'mg', 'ms', 'ml', 'mt', 'mi', 'mr', 'mn', 'my', 'ne', 'no', 'or',
        'ps', 'fa', 'pl', 'pt', 'pa', 'ro', 'ru', 'sm', 'gd', 'sr', 'st', 'sn',
        'sd', 'si', 'sk', 'sl', 'so', 'es', 'su', 'sw', 'sv', 'tg', 'ta', 'tt',
        'te', 'th', 'tr', 'tk', 'uk', 'ur', 'ug', 'uz', 'vi', 'cy', 'xh', 'yi',
        'yo', 'zu'
    ]
    chars_limit = 2000
    supported_features = {
        'detection': True,
        'mistakes': True,
        'pronunciation': True,
        'change-instance': False,
        'suggestions': False,
        'api-key-supported': False,
        'api-key-required': False,
    }

    _service_urls = DEFAULT_SERVICE_URLS
    _headers = {
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64)',
        'Referer': 'https://translate.google.com',
    }
    _src_lang = None
    _dest_lang = None

    def __init__(self, **kwargs):
        pass

    @staticmethod
    def _build_rpc_request(text: str, src: str, dest: str):
        return json.dumps([[
            [
                RPC_ID,
                json.dumps([[text, src, dest, True], [None]], separators=(',', ':')),
                None,
                'generic',
            ],
        ]], separators=(',', ':'))

    def _pick_service_url(self):
        if len(self._service_urls) == 1:
            return self._service_urls[0]
        return random.choice(self._service_urls)

    @property
    def translate_url(self):
        url = TRANSLATE_RPC.format(host=self._pick_service_url()) + '?'
        params_dict = {
            'rpcids': RPC_ID,
            'bl': 'boq_translate-webserver_20201207.13_p0',
            'soc-app': '1',
            'soc-platform': '1',
            'soc-device': '1',
            'rt': 'c',
        }
        params = []
        for item in params_dict:
            params.append('='.join([item, params_dict[item]]))

        return self.format_url(url, '&'.join(params))

    def format_translation(self, text, src, dest):
        data = {
            'f.req': self._build_rpc_request(text, src, dest),
        }
        self._src_lang = src
        self._dest_lang = dest
        return ('POST', data, self._headers, True)

    def get_translation(self, data):
        try:
            token_found = False
            square_bracket_counts = [0, 0]
            resp = ''
            data = data.decode('utf-8')

            for line in data.split('\n'):
                token_found = token_found or f'"{RPC_ID}"' in line[:30]
                if not token_found:
                    continue

                is_in_string = False
                for index, char in enumerate(line):
                    if char == '\"' and line[max(0, index - 1)] != '\\':
                        is_in_string = not is_in_string
                    if not is_in_string:
                        if char == '[':
                            square_bracket_counts[0] += 1
                        elif char == ']':
                            square_bracket_counts[1] += 1

                resp += line
                if square_bracket_counts[0] == square_bracket_counts[1]:
                    break

            data = json.loads(resp)
            parsed = json.loads(data[0][2])
            should_spacing = parsed[1][0][0][3]
            translated_parts = None
            translated = None
            try:
                translated_parts = list(
                    map(
                        lambda part: TranslatedPart(
                            part[0] if len(part) > 0 else '',
                            part[1] if len(part) >= 2 else []
                        ),
                        parsed[1][0][0][5]
                    )
                )
            except TypeError:
                translated_parts = [
                    TranslatedPart(
                        parsed[1][0][1][0],
                        [parsed[1][0][0][0], parsed[1][0][1][0]]
                    )
                ]
            translated = (' ' if should_spacing else '').join(map(lambda part: part.text, translated_parts))

            src = None
            try:
                src = parsed[1][-1][1]
            except (IndexError, TypeError):
                pass

            if not src == self._src_lang:
                raise TranslationError('source language mismatch')

            if src == 'auto':
                try:
                    if parsed[0][2] in self.languages:
                        src = parsed[0][2]
                except (IndexError, TypeError):
                    pass

            dest = None
            try:
                dest = parsed[1][-1][2]
            except (IndexError, TypeError):
                pass

            if not dest == self._dest_lang:
                raise TranslationError('destination language mismatch')

            origin_pronunciation = None
            try:
                origin_pronunciation = parsed[0][0]
            except (IndexError, TypeError):
                pass

            pronunciation = None
            try:
                pronunciation = parsed[1][0][0][1]
            except (IndexError, TypeError):
                pass

            mistake = None
            try:
                mistake = parsed[0][1][0][0][1]
            except (IndexError, TypeError):
                pass

            result = Translation(
                translated,
                {
                    'possible-mistakes': [mistake, self._strip_html_tags(mistake)],
                    'src-pronunciation': origin_pronunciation,
                    'dest-pronunciation': pronunciation,
                }
            )
            return (result, src)
        except TranslationError as e:
            raise e
        except Exception as e:
            raise TranslatorError(str(e))

    def _strip_html_tags(self, text):
        """Strip html tags"""
        if text is None:
            return None

        tags_re = re.compile(r'(<!--.*?-->|<[^>]*>)')
        tags_removed = tags_re.sub('', text)
        escaped = html.escape(tags_removed)
        return escaped


class TranslatedPart:
    def __init__(self, text: str, candidates: List[str]):
        self.text = text
        self.candidates = candidates

    def __str__(self):
        return self.text

    def __dict__(self):
        return {
            'text': self.text,
            'candidates': self.candidates,
        }



# Copyright 2020-2022 Mufeed Ali
# Copyright 2020-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import re

from gi.repository import Gio, Gdk, GObject, Gtk

from dialect.define import RES_PATH
from dialect.translators import get_lang_name


@Gtk.Template(resource_path=f'{RES_PATH}/lang-selector.ui')
class DialectLangSelector(Gtk.Popover):
    __gtype_name__ = 'DialectLangSelector'
    __gsignals__ = {
        'user-selection-changed': (GObject.SIGNAL_RUN_LAST, GObject.TYPE_NONE, ())
    }

    # Get widgets
    search = Gtk.Template.Child()
    scroll = Gtk.Template.Child()
    revealer = Gtk.Template.Child()
    recent_list = Gtk.Template.Child()
    separator = Gtk.Template.Child()
    lang_list = Gtk.Template.Child()

    # Propeties
    selected = GObject.Property(type=str)  # Key of the selected lang

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Connect popover open/close signal
        self.connect('show', self._show)
        self.connect('closed', self._closed)

        # Connect list signals
        self.recent_list.connect('row-activated', self._activated)
        self.lang_list.connect('row-activated', self._activated)

        # Setup search entry
        self.search.set_key_capture_widget(self)
        key_events = Gtk.EventControllerKey.new()
        key_events.connect('key-pressed', self._on_key_pressed)
        self.search.add_controller(key_events)
        # Connect search entry signals
        self.search.connect('changed', self._on_search)
        self.search.connect('activate', self._on_search_activate)

        self.recent_model = Gio.ListStore.new(LangObject)
        self.recent_list.bind_model(self.recent_model, self._create_lang_row)

        self.lang_model = Gio.ListStore.new(LangObject)
        self.filter = Gtk.CustomFilter()
        self.filter.set_filter_func(self._filter_func)
        fitler_model = Gtk.FilterListModel.new(self.lang_model, self.filter)
        self.lang_list.bind_model(fitler_model, self._create_lang_row)

    def get_selected(self):
        return self.get_property('selected')

    def set_selected(self, lang_code, notify=True):
        self.set_property('selected', lang_code)
        if notify:
            self.emit('user-selection-changed')

    def set_languages(self, languages):
        # Clear list
        self.lang_model.remove_all()

        # Load langs list
        for code in languages:
            self.lang_model.append(LangObject(code, get_lang_name(code)))

    def insert_recent(self, code, name):
        row_selected = (code == self.selected)
        self.recent_model.append(LangObject(code, name, row_selected))

    def clear_recent(self):
        self.recent_model.remove_all()

    def refresh_selected(self):
        for item in self.lang_model:
            item.set_property('selected', (item.code == self.selected))

    def _activated(self, _list, row):
        # Close popover
        self.popdown()
        # Set selected property
        self.set_selected(row.lang.code)

    def _show(self, _popover):
        self.search.grab_focus()

    def _closed(self, _popover):
        # Reset scroll
        vscroll = self.scroll.get_vadjustment()
        vscroll.set_value(0)
        # Clear search
        self.search.set_text('')

    def _create_lang_row(self, lang):
        return DialectLangRow(lang)

    def _filter_func(self, item):
        search = self.search.get_text()
        return bool(re.search(search, item.name, re.IGNORECASE))

    def _on_search(self, _entry):
        search = self.search.get_text()
        if search != '':
            self.revealer.set_reveal_child(False)
        else:
            self.revealer.set_reveal_child(True)

        self.filter.emit('changed', Gtk.FilterChange.DIFFERENT)

    def _on_search_activate(self, _entry):
        if self.search.get_text():
            row = self.lang_list.get_row_at_index(0)
            if row:
                self.lang_list.emit('row-activated', row)
        return Gdk.EVENT_PROPAGATE

    def _on_key_pressed(self, _controller, keyval, _keycode, _mod):
        # Close popover if ESQ key is pressed in search entry
        if keyval == Gdk.KEY_Escape:
            self.popdown()
        # Prevent search entry unfocusing when down key is pressed
        elif keyval == Gdk.KEY_Down:
            return Gdk.EVENT_STOP


class LangObject(GObject.Object):
    __gtype_name__ = 'LangObject'

    code = GObject.Property(type=str)
    name = GObject.Property(type=str)
    selected = GObject.Property(type=bool, default=False)

    def __init__(self, code, name, selected=False):
        super().__init__()

        self.set_property('code', code)
        self.set_property('name', name)
        self.set_property('selected', selected)


@Gtk.Template(resource_path=f'{RES_PATH}/lang-row.ui')
class DialectLangRow(Gtk.ListBoxRow):
    __gtype_name__ = 'DialectLangRow'

    # Widgets
    name = Gtk.Template.Child()
    selection = Gtk.Template.Child()

    def __init__(self, lang):
        super().__init__()
        self.lang = lang
        self.name.set_label(self.lang.name)

        self.lang.bind_property(
            'selected',
            self.selection,
            'visible',
            GObject.BindingFlags.SYNC_CREATE
        )



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gio, GLib

from dialect.define import APP_ID
from dialect.translators import (
    check_backend_availability,
    get_fallback_backend_name,
    TRANSLATORS
)
from dialect.tts import TTS


class Settings(Gio.Settings):
    """
    Dialect settings handler
    """

    instance = None
    translator = None

    def __init__(self):
        Gio.Settings.__init__(self)

    @staticmethod
    def new():
        """Create a new instance of Settings."""
        g_settings = Gio.Settings.new(APP_ID)
        g_settings.__class__ = Settings
        g_settings.init_translators_settings()
        return g_settings

    @staticmethod
    def get():
        """Return an active instance of Settings."""
        if Settings.instance is None:
            Settings.instance = Settings.new()
        return Settings.instance

    def init_translators_settings(self):
        """Intialize translators settings with its default values."""
        self.translators_list = list(TRANSLATORS.keys())
        for name, instance in TRANSLATORS.items():
            settings = self.get_translator_settings(name)
            if not settings.get_boolean('init'):
                settings.set_strv('src-langs', instance.src_langs)
                settings.set_strv('dest-langs', instance.dest_langs)
                settings.set_string('instance-url', instance.instance_url)
                settings.set_string('api-key', instance.api_key)
                settings.set_boolean('init', True)

    @property
    def translators_list(self):
        return self.get_child('translators').get_strv('list')

    @translators_list.setter
    def translators_list(self, translators):
        self.get_child('translators').set_strv('list', translators)

    @property
    def active_translator(self):
        value = self.get_child('translators').get_string('active')

        if check_backend_availability(value):
            return value

        self.active_translator = get_fallback_backend_name()
        return get_fallback_backend_name()

    @active_translator.setter
    def active_translator(self, translator):
        self.get_child('translators').set_string('active', translator)
        self.translator = None

    def get_translator_settings(self, translator=None):
        def on_changed(_settings, key):
            self.emit('changed', 'translator-' + key)

        def get_settings(name):
            path = self.get_child('translators').get_property('path')
            if not path.endswith('/'):
                path += '/'
            path += name + '/'
            settings = Gio.Settings(APP_ID + '.translator', path)
            settings.connect('changed', on_changed)
            return settings

        if translator is not None:
            translator = get_settings(translator)
            return translator
        if self.translator is None:
            self.translator = get_settings(self.active_translator)
            self.translator.delay()
        return self.translator

    def save_translator_settings(self):
        if self.translator is not None:
            self.translator.apply()

    @property
    def src_langs(self):
        return self.get_translator_settings().get_strv('src-langs')

    @src_langs.setter
    def src_langs(self, src_langs):
        self.get_translator_settings().set_strv('src-langs', src_langs)

    def reset_src_langs(self):
        self.src_langs = TRANSLATORS[self.active_translator].src_langs

    @property
    def dest_langs(self):
        return self.get_translator_settings().get_strv('dest-langs')

    @dest_langs.setter
    def dest_langs(self, dest_langs):
        self.get_translator_settings().set_strv('dest-langs', dest_langs)

    def reset_dest_langs(self):
        self.dest_langs = TRANSLATORS[self.active_translator].dest_langs

    @property
    def instance_url(self):
        return self.get_translator_settings().get_string('instance-url')

    @instance_url.setter
    def instance_url(self, url):
        self.get_translator_settings().set_string('instance-url', url)
        self.save_translator_settings()

    def reset_instance_url(self):
        self.instance_url = TRANSLATORS[self.active_translator].instance_url

    @property
    def api_key(self):
        return self.get_translator_settings().get_string('api-key')

    @api_key.setter
    def api_key(self, api_key):
        self.get_translator_settings().set_string('api-key', api_key)
        self.save_translator_settings()

    def reset_api_key(self):
        self.api_key = TRANSLATORS[self.active_translator].api_key

    @property
    def window_size(self):
        value = self.get_value('window-size')
        return (value[0], value[1])

    @window_size.setter
    def window_size(self, size):
        width, height = size
        self.set_value('window-size', GLib.Variant('ai', [width, height]))

    @property
    def translate_accel(self):
        """Return the user's preferred translation shortcut."""
        value = self.translate_accel_value

        if value == 0:
            return '<Primary>Return'
        if value == 1:
            return 'Return'

        return '<Primary>Return'

    @property
    def translate_accel_value(self):
        """Return the user's preferred translation shortcut value."""
        return self.get_int('translate-accel')

    @property
    def active_tts(self):
        """Return the user's preferred TTS service."""
        value = self.get_child('tts').get_string('active')

        if value != '' and value not in TTS.keys():
            value = ''
            self.active_tts = value

        return value

    @active_tts.setter
    def active_tts(self, tts):
        """Set the user's preferred TTS service."""
        self.get_child('tts').set_string('active', tts)

    @property
    def color_scheme(self):
        return self.get_string('color-scheme')

    @color_scheme.setter
    def color_scheme(self, scheme):
        self.set_string('dark-mode', scheme)

    @property
    def live_translation(self):
        return self.get_boolean('live-translation')

    @live_translation.setter
    def live_translation(self, state):
        self.set_boolean('live-translation', state)

    @property
    def sp_translation(self):
        return self.get_boolean('sp-translation')

    @sp_translation.setter
    def sp_translation(self, state):
        self.set_boolean('sp-translation', state)

    @property
    def show_pronunciation(self):
        return self.get_boolean('show-pronunciation')

    @property
    def show_pronunciation_value(self):
        return self.get_value('show-pronunciation')

    @show_pronunciation.setter
    def show_pronunciation(self, state):
        self.set_boolean('show-pronunciation', state)

    @property
    def src_auto(self):
        return self.get_boolean('src-auto')

    @src_auto.setter
    def src_auto(self, state):
        self.set_boolean('src-auto', state)



# Copyright 2020 Manuel Genovés
# Copyright 2022 Mufeed Ali
# Copyright 2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

# Code modifed from Apostrophe
# https://gitlab.gnome.org/World/apostrophe/-/blob/main/apostrophe/theme_switcher.py

from gi.repository import Adw, Gio, GObject, Gtk 

from dialect.define import RES_PATH
from dialect.settings import Settings


@Gtk.Template(resource_path=f'{RES_PATH}/theme-switcher.ui')
class DialectThemeSwitcher(Gtk.Box):
    __gtype_name__ = 'DialectThemeSwitcher'

    show_system = GObject.property(type=bool, default=True)
    color_scheme = 'light'

    system = Gtk.Template.Child()
    light = Gtk.Template.Child()
    dark = Gtk.Template.Child()

    @GObject.Property(type=str)
    def selected_color_scheme(self):
        """Read-write integer property."""

        return self.color_scheme

    @selected_color_scheme.setter
    def selected_color_scheme(self, color_scheme):
        self.color_scheme = color_scheme

        if color_scheme == 'auto':
            self.system.set_active(True)
            self.style_manager.set_color_scheme(Adw.ColorScheme.PREFER_LIGHT)
        if color_scheme == 'light':
            self.light.set_active(True)
            self.style_manager.set_color_scheme(Adw.ColorScheme.FORCE_LIGHT)
        if color_scheme == 'dark':
            self.dark.set_active(True)
            self.style_manager.set_color_scheme(Adw.ColorScheme.FORCE_DARK)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.style_manager = Adw.StyleManager.get_default()

        self.color_scheme = Settings.get().color_scheme

        Settings.get().bind(
            'color-scheme',
            self,
            'selected_color_scheme',
            Gio.SettingsBindFlags.DEFAULT
        )

        self.style_manager.bind_property(
            'system-supports-color-schemes',
            self, 'show_system',
            GObject.BindingFlags.SYNC_CREATE
        )

    @Gtk.Template.Callback()
    def _on_color_scheme_changed(self, _widget, _paramspec):
        if self.system.get_active():
            self.selected_color_scheme = 'auto'
        if self.light.get_active():
            self.selected_color_scheme = 'light'
        if self.dark.get_active():
            self.selected_color_scheme = 'dark'



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

from gi.repository import Gtk

from dialect.define import RES_PATH
from dialect.settings import Settings


@Gtk.Template(resource_path=f'{RES_PATH}/shortcuts.ui')
class DialectShortcutsWindow(Gtk.ShortcutsWindow):
    __gtype_name__ = 'DialectShortcutsWindow'

    translate_shortcut = Gtk.Template.Child()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.connect('show', self._on_show)

    def _on_show(self, _data):
        self.translate_shortcut.set_visible(not Settings.get().live_translation)
        self.translate_shortcut.set_property('accelerator', Settings.get().translate_accel)



# Copyright 2021 Mufeed Ali
# Copyright 2021 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from gi.repository import Soup

from dialect.translators.basetrans import (
    InvalidLangCode, TranslatorBase, Translation, TranslatorError
)
from dialect.session import Session, ResponseEmpty


class Translator(TranslatorBase):
    name = 'lingva'
    prettyname = 'Lingva Translate'
    history = []
    languages = []
    chars_limit = 5000
    supported_features = {
        'detection': True,
        'mistakes': False,
        'pronunciation': False,
        'change-instance': True,
        'suggestions': False,
        'api-key-supported': False,
        'api-key-required': False,
    }
    instance_url = 'lingva.ml'

    validation_path = '/api/v1/en/es/hello'

    def __init__(self, callback, base_url=None, _api_key='', **kwargs):
        def on_langs_response(session, result):
            success = False
            network_error = False
            error = ''
            try:
                data = Session.get_response(session, result)
                self._check_errors(data)
                if 'languages' in data:
                    for lang in data['languages']:
                        if lang['code'] != 'auto':
                            self.languages.append(lang['code'])
                    success = True
            except (TranslatorError, ResponseEmpty) as exc:
                logging.warning(exc)
                error = str(exc)
            except Exception as exc:
                logging.warning(exc)
                error = str(exc)
                network_error = True

            callback(success, error, network_error)

        if base_url is not None:
            self.instance_url = base_url

        lang_message = Soup.Message.new('GET', self.lang_url)
        Session.get().send_and_read_async(lang_message, 0, None, on_langs_response)

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, '/api/v1/languages/')

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, '/api/v1/{src}/{dest}/{text}')

    @staticmethod
    def validate_instance(data):
        valid = False

        if data and 'translation' in data:
            valid = True

        return valid

    def format_translation(self, _text, _src, _dest):
        return ('GET', {}, {}, False)

    def get_translation(self, data):
        self._check_errors(data)
        translation = Translation(
            data['translation'],
            {
                'possible-mistakes': None,
                'src-pronunciation': None,
                'dest-pronunciation': None,
            },
        )

        return (translation, None)

    def _check_errors(self, data):
        """Raises a proper Exception if an error is found in the data."""
        if 'error' in data:
            error = data['error']

            if error == 'Invalid target language' or error == 'Invalid source language':
                raise InvalidLangCode(error)
            else:
                raise TranslatorError(error)



# Copyright 2021-2022 Mufeed Ali
# Copyright 2021-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging

from gi.repository import Soup

from dialect.translators.basetrans import (
    ApiKeyRequired, BatchSizeExceeded, CharactersLimitExceeded,
    InvalidLangCode, InvalidApiKey, TranslatorBase, Translation,
    TranslationError, TranslatorError
)
from dialect.session import Session, ResponseEmpty


class Translator(TranslatorBase):
    name = 'libretranslate'
    prettyname = 'LibreTranslate'
    history = []
    languages = []
    chars_limit = 0
    supported_features = {
        'detection': True,
        'mistakes': False,
        'pronunciation': False,
        'change-instance': True,
        'suggestions': False,
        'api-key-supported': False,
        'api-key-required': False,
    }
    instance_url = 'libretranslate.de'
    api_key = ''

    validation_path = '/spec'
    api_test_path = '/translate'

    def __init__(self, callback, base_url=None, api_key='', **kwargs):
        def on_loaded():
            callback(self.langs_success and self.settings_success, self.error, self.network_error)

        def on_langs_response(session, result):
            try:
                data = Session.get_response(session, result)
                self._check_errors(data)
                for lang in data:
                    self.languages.append(lang['code'])
                self.langs_success = True
            except (TranslatorError, ResponseEmpty) as exc:
                logging.warning(exc)
                self.error = str(exc)
            except Exception as exc:
                logging.warning(exc)
                self.error = str(exc)
                self.network_error = True

        def on_settings_response(session, result):
            try:
                data = Session.get_response(session, result)
                self._check_errors(data)
                self.supported_features['suggestions'] = data.get('suggestions', False)
                self.supported_features['api-key-supported'] = data.get('apiKeys', False)
                self.supported_features['api-key-required'] = data.get('keyRequired', False)
                self.chars_limit = data.get('charLimit', 0)
                self.settings_success = True
            except (TranslatorError, ResponseEmpty) as exc:
                logging.warning(exc)
                self.error = str(exc)
            except Exception as exc:
                logging.warning(exc)
                self.error = str(exc)
                self.network_error = True

        self.langs_success = False
        self.settings_success = False
        self.network_error = False
        self.error = ''

        if base_url is not None:
            self.instance_url = base_url
        self.api_key = api_key

        lang_message = Soup.Message.new('GET', self.lang_url)
        settings_message = Soup.Message.new('GET', self._frontend_settings_url)

        Session.get().multiple(
            [
                [lang_message, on_langs_response],
                [settings_message, on_settings_response]
            ],
            on_loaded
        )

    @property
    def _frontend_settings_url(self):
        return self.format_url(self.instance_url, '/frontend/settings')

    @property
    def detect_url(self):
        return self.format_url(self.instance_url, '/detect')

    @property
    def lang_url(self):
        return self.format_url(self.instance_url, '/languages')

    @property
    def suggest_url(self):
        return self.format_url(self.instance_url, '/suggest')

    @property
    def translate_url(self):
        return self.format_url(self.instance_url, '/translate')

    @staticmethod
    def validate_instance(data):
        valid = False

        if data and data is not None:
            valid = data['info']['title'] == 'LibreTranslate'

        return valid

    @staticmethod
    def format_api_key_test(api_key):
        data = {
            'q': 'hello',
            'source': 'en',
            'target': 'es',
            'api_key': api_key,
        }

        return (data, {})

    def format_suggestion(self, text, src, dest, suggestion):
        data = {
            'q': text,
            'source': src,
            'target': dest,
            's': suggestion,
        }
        if self.api_key:
            data['api_key'] = self.api_key

        return (data, {})

    def get_suggestion(self, data):
        self._check_errors(data)
        return data.get('success', False)

    def format_translation(self, text, src, dest):
        data = {
            'q': text,
            'source': src,
            'target': dest,
        }
        if self.api_key:
            data['api_key'] = self.api_key

        return ('POST', data, {}, False)

    def get_translation(self, data):
        self._check_errors(data)
        translation = Translation(
            data['translatedText'],
            {
                'possible-mistakes': None,
                'src-pronunciation': None,
                'dest-pronunciation': None,
            },
        )

        return (translation, None)

    def _check_errors(self, data):
        """Raises a proper Exception if an error is found in the data."""
        if 'error' in data:
            error = data['error']

            if error == 'Please contact the server operator to obtain an API key':
                raise ApiKeyRequired(error)
            elif error == 'Invalid API key':
                raise InvalidApiKey(error)
            elif 'is not supported' in error:
                raise InvalidLangCode(error)
            elif 'exceeds text limit' in error:
                raise BatchSizeExceeded(error)
            elif 'exceeds character limit' in error:
                raise CharactersLimitExceeded(error)
            elif 'Cannot translate text' in error or 'format is not supported' in error:
                raise TranslationError(error)
            else:
                raise TranslatorError(error)



import importlib
import pkgutil

TTS = {}
for _importer, modname, _ispkg in pkgutil.iter_modules(__path__):
    if modname != 'basetts':
        modclass = importlib.import_module('dialect.tts.' + modname).TextToSpeech
        TTS[modclass.name] = modclass



# Copyright 2020 gi-lom
# Copyright 2020-2022 Mufeed Ali
# Copyright 2020-2022 Rafael Mardojai CM
# SPDX-License-Identifier: GPL-3.0-or-later

import logging
import threading
from gettext import gettext as _
from tempfile import NamedTemporaryFile

from gi.repository import Adw, Gdk, Gio, GLib, GObject, Gst, Gtk

from dialect.define import APP_ID, PROFILE, RES_PATH, TRANS_NUMBER
from dialect.lang_selector import DialectLangSelector
from dialect.session import Session, ResponseEmpty, ResponseError
from dialect.settings import Settings
from dialect.shortcuts import DialectShortcutsWindow
from dialect.theme_switcher import DialectThemeSwitcher
from dialect.translators import TRANSLATORS, get_lang_name
from dialect.translators.basetrans import ApiKeyRequired, InvalidApiKey, TranslatorError
from dialect.tts import TTS


@Gtk.Template(resource_path=f'{RES_PATH}/window.ui')
class DialectWindow(Adw.ApplicationWindow):
    __gtype_name__ = 'DialectWindow'

    # Get widgets
    menu_btn = Gtk.Template.Child()
    main_stack = Gtk.Template.Child()
    error_page = Gtk.Template.Child()
    translator_box = Gtk.Template.Child()
    key_page = Gtk.Template.Child()
    rmv_key_btn = Gtk.Template.Child()
    error_api_key_btn = Gtk.Template.Child()

    title_stack = Gtk.Template.Child()
    langs_button_box = Gtk.Template.Child()
    switch_btn = Gtk.Template.Child()
    src_lang_btn = Gtk.Template.Child()
    src_lang_label = Gtk.Template.Child()
    dest_lang_btn = Gtk.Template.Child()
    dest_lang_label = Gtk.Template.Child()

    return_btn = Gtk.Template.Child()
    forward_btn = Gtk.Template.Child()

    src_pron_revealer = Gtk.Template.Child()
    src_pron_label = Gtk.Template.Child()
    mistakes = Gtk.Template.Child()
    mistakes_label = Gtk.Template.Child()
    char_counter = Gtk.Template.Child()
    src_text = Gtk.Template.Child()
    clear_btn = Gtk.Template.Child()
    paste_btn = Gtk.Template.Child()
    src_voice_btn = Gtk.Template.Child()
    translate_btn = Gtk.Template.Child()

    dest_box = Gtk.Template.Child()
    dest_pron_revealer = Gtk.Template.Child()
    dest_pron_label = Gtk.Template.Child()
    dest_text = Gtk.Template.Child()
    dest_toolbar_stack = Gtk.Template.Child()
    trans_spinner = Gtk.Template.Child()
    trans_warning = Gtk.Template.Child()
    edit_btn = Gtk.Template.Child()
    copy_btn = Gtk.Template.Child()
    dest_voice_btn = Gtk.Template.Child()

    actionbar = Gtk.Template.Child()
    src_lang_btn2 = Gtk.Template.Child()
    switch_btn2 = Gtk.Template.Child()
    dest_lang_btn2 = Gtk.Template.Child()

    toast = None  # for notification management
    toast_overlay = Gtk.Template.Child()

    src_key_ctrlr = Gtk.Template.Child()
    win_key_ctrlr = Gtk.Template.Child()

    # Translator
    translator = None  # Translator object
    # Text to speech
    tts = None
    tts_langs = None
    voice_loading = False  # tts loading status

    # Preset language values
    src_langs = []
    dest_langs = []

    current_history = 0  # for history management

    # Translation-related variables
    next_trans = {}  # for ongoing translation
    ongoing_trans = False  # for ongoing translation
    trans_failed = False  # for monitoring connectivity issues
    trans_mistakes = [None, None]  # "mistakes" suggestions
    # Pronunciations
    trans_src_pron = None
    trans_dest_pron = None
    # Suggestions
    before_suggest = None

    mobile_mode = False  # UI mode

    # Propeties
    backend_loading = GObject.Property(type=bool, default=False)

    def __init__(self, text, langs, **kwargs):
        super().__init__(**kwargs)

        # Options passed to command line
        self.launch_text = text
        self.launch_langs = langs

        # Application object
        self.app = kwargs['application']

        # GStreamer playbin object and related setup
        self.player = Gst.ElementFactory.make('playbin', 'player')
        bus = self.player.get_bus()
        bus.add_signal_watch()
        bus.connect('message', self.on_gst_message)
        self.player_event = threading.Event()  # An event for letting us know when Gst is done playing

        # Setup window
        self.setup_actions()
        self.setup()

    def setup(self):
        self.set_default_icon_name(APP_ID)

        # Set devel style
        if PROFILE == 'Devel':
            self.get_style_context().add_class('devel')

        # Theme Switcher
        theme_switcher = DialectThemeSwitcher()
        self.menu_btn.get_popover().add_child(theme_switcher, 'theme')

        # Save settings on close
        self.connect('unrealize', self.save_settings)

        self.setup_headerbar()
        self.setup_translation()
        self.set_help_overlay(DialectShortcutsWindow())

        # Load translator
        self.load_translator(True)
        # Get languages available for speech
        if Settings.get().active_tts != '':
            threading.Thread(target=self.load_lang_speech, daemon=True).start()

    def do_size_allocate(self, width, height, baseline):
        if width < 680:
            if self.mobile_mode is False:
                self.mobile_mode = True
                self.toggle_mobile_mode()
        else:
            if self.mobile_mode is True:
                self.mobile_mode = False
                self.toggle_mobile_mode()

        Adw.ApplicationWindow.do_size_allocate(self, width, height, baseline)

    def setup_actions(self):
        back = Gio.SimpleAction.new('back', None)
        back.set_enabled(False)
        back.connect('activate', self.ui_return)
        self.add_action(back)

        forward_action = Gio.SimpleAction.new('forward', None)
        forward_action.set_enabled(False)
        forward_action.connect('activate', self.ui_forward)
        self.add_action(forward_action)

        switch_action = Gio.SimpleAction.new('switch', None)
        switch_action.connect('activate', self.ui_switch)
        self.add_action(switch_action)

        clear_action = Gio.SimpleAction.new('clear', None)
        clear_action.set_enabled(False)
        clear_action.connect('activate', self.ui_clear)
        self.add_action(clear_action)

        paste_action = Gio.SimpleAction.new('paste', None)
        paste_action.connect('activate', self.ui_paste)
        self.add_action(paste_action)

        copy_action = Gio.SimpleAction.new('copy', None)
        copy_action.set_enabled(False)
        copy_action.connect('activate', self.ui_copy)
        self.add_action(copy_action)

        listen_dest_action = Gio.SimpleAction.new('listen-dest', None)
        listen_dest_action.connect('activate', self.ui_dest_voice)
        self.add_action(listen_dest_action)

        suggest_action = Gio.SimpleAction.new('suggest', None)
        suggest_action.set_enabled(False)
        suggest_action.connect('activate', self.ui_suggest)
        self.add_action(suggest_action)

        suggest_ok_action = Gio.SimpleAction.new('suggest-ok', None)
        suggest_ok_action.connect('activate', self.ui_suggest_ok)
        self.add_action(suggest_ok_action)

        suggest_cancel_action = Gio.SimpleAction.new('suggest-cancel', None)
        suggest_cancel_action.connect('activate', self.ui_suggest_cancel)
        self.add_action(suggest_cancel_action)

        listen_src_action = Gio.SimpleAction.new('listen-src', None)
        listen_src_action.connect('activate', self.ui_src_voice)
        self.add_action(listen_src_action)

        translation_action = Gio.SimpleAction.new('translation', None)
        translation_action.set_enabled(False)
        translation_action.connect('activate', self.translation)
        self.add_action(translation_action)

    def load_translator(self, launch=False):
        def on_loaded(success, error='', network_error=False):
            if success:
                # Supported features
                if not self.translator.supported_features['mistakes']:
                    self.mistakes.set_reveal_child(False)

                self.ui_suggest_cancel(None, None)
                if not self.translator.supported_features['suggestions']:
                    self.edit_btn.set_visible(False)
                else:
                    self.edit_btn.set_visible(True)

                if not self.translator.supported_features['pronunciation']:
                    self.src_pron_revealer.set_reveal_child(False)
                    self.dest_pron_revealer.set_reveal_child(False)
                    self.app.lookup_action('pronunciation').set_enabled(False)
                else:
                    self.app.lookup_action('pronunciation').set_enabled(True)

                # Update langs list
                self.src_lang_selector.set_languages(self.translator.languages)
                self.dest_lang_selector.set_languages(self.translator.languages)
                # Update selected langs
                set_auto = Settings.get().src_auto and self.translator.supported_features['detection']
                self.src_lang_selector.set_selected(
                    'auto' if set_auto else self.src_langs[0],
                    notify=False
                )
                self.dest_lang_selector.set_selected(self.dest_langs[0], notify=False)
                # Update chars limit
                self.char_counter.set_text(
                    f'{str(self.src_buffer.get_char_count())}/{self.translator.chars_limit}'
                )

                self.set_property('backend-loading', False)

                self.check_apikey()
            else:
                # Show error view
                self.loading_failed(error, network_error)
                self.set_property('backend-loading', False)

        backend = Settings.get().active_translator

        # Get saved languages
        self.src_langs = Settings.get().src_langs
        self.dest_langs = Settings.get().dest_langs

        # Show loading view
        self.main_stack.set_visible_child_name('loading')

        # Translator object
        self.translator = TRANSLATORS[backend](
            callback=on_loaded,
            base_url=Settings.get().instance_url,
            api_key=Settings.get().api_key,
        )
        if not TRANSLATORS[backend].supported_features['change-instance']:
            on_loaded(True)  # Assume successful loading.

        if launch:
            if self.launch_langs['src'] is not None:
                self.src_lang_selector.set_selected(self.launch_langs['src'], notify=False)
            if self.launch_langs['dest'] is not None and self.launch_langs['dest'] in self.translator.languages:
                self.dest_lang_selector.set_selected(self.launch_langs['dest'], notify=False)

            if self.launch_text != '':
                self.translate(self.launch_text, self.launch_langs['src'], self.launch_langs['dest'])

    def check_apikey(self):
        def on_response(session, result):
            try:
                data = Session.get_response(session, result)
                self.translator.get_translation(data)
                self.main_stack.set_visible_child_name('translate')
            except InvalidApiKey as exc:
                logging.warning(exc)
                self.key_page.set_title(_('The provided API key is invalid'))
                if self.translator.supported_features['api-key-required']:
                    self.key_page.set_description(_('Please set a valid API key in the preferences.'))
                else:
                    self.key_page.set_description(
                        _('Please set a valid API key or unset the API key in the preferences.')
                    )
                    self.rmv_key_btn.set_visible(True)
                    self.error_api_key_btn.set_visible(True)
                self.main_stack.set_visible_child_name('api-key')
            except (TranslatorError, ResponseEmpty) as exc:
                logging.warning(exc)
                self.loading_failed(str(exc))
            except Exception as exc:
                logging.warning(exc)
                self.loading_failed(str(exc), True)

        if self.translator.supported_features['api-key-supported']:
            if Settings.get().api_key:
                validation_url = self.translator.format_url(
                    Settings.get().instance_url,
                    self.translator.api_test_path
                )
                (data, headers) = self.translator.format_api_key_test(Settings.get().api_key)
                message = Session.create_message('POST', validation_url, data, headers)
                Session.get().send_and_read_async(message, 0, None, on_response)
            elif not Settings.get().api_key and self.translator.supported_features['api-key-required']:
                self.key_page.set_title(_('API key is required to use the service'))
                self.key_page.set_description(_('Please set an API key in the preferences.'))
                self.main_stack.set_visible_child_name('api-key')
            else:
                self.main_stack.set_visible_child_name('translate')
        else:
            self.main_stack.set_visible_child_name('translate')

    def loading_failed(self, details='', network=False):
        self.main_stack.set_visible_child_name('error')

        service = self.translator.prettyname
        url = self.translator.instance_url

        title = _('Failed loading the translation service')
        description = _('Please report this in the Dialect bug tracker if the issue persists.')
        if self.translator.supported_features['change-instance']:
            description = _('Failed loading "{url}", check if the instance address is correct or report in the Dialect bug tracker if the issue persists.')
            description = description.format(url=url)

        if network:
            title = _('Couldn’t connect to the translation service')
            description = _('We can’t connect to the server. Please check for network issues.')
            if self.translator.supported_features['change-instance']:
                description = _('We can’t connect to the {service} instance "{url}".\n Please check for network issues or if the address is correct.')
                description = description.format(service=service, url=url)

        if details:
            description = description + '\n\n<small><tt>' + details + '</tt></small>'

        self.error_page.set_title(title)
        self.error_page.set_description(description)

    @Gtk.Template.Callback()
    def retry_load_translator(self, _button):
        self.load_translator()

    @Gtk.Template.Callback()
    def remove_key_and_reload(self, _button):
        Settings.get().reset_api_key()
        self.load_translator()

    def on_listen_failed(self, called_from):
        self.src_voice_btn.set_child(self.src_voice_warning)
        self.src_voice_spinner.stop()

        self.dest_voice_btn.set_child(self.dest_voice_warning)
        self.dest_voice_spinner.stop()

        tooltip_text = _('A network issue has occured. Retry?')
        self.src_voice_btn.set_tooltip_text(tooltip_text)
        self.dest_voice_btn.set_tooltip_text(tooltip_text)

        if called_from is not None:
            action = {
                'label': _('Retry'),
                'name': 'win.listen-src' if called_from == 'src' else 'win.listen-dest',
            }
        else:
            action = None

        self.send_notification(
            _('A network issue has occured. Please try again.'),
            action=action
        )

        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )

        if self.tts_langs:
            self.lookup_action('listen-src').set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and src_text != ''
            )
            self.lookup_action('listen-dest').set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and dest_text != ''
            )
        else:
            self.lookup_action('listen-src').set_enabled(src_text != '')
            self.lookup_action('listen-dest').set_enabled(dest_text != '')

    def load_lang_speech(self, listen=False, text=None, language=None, called_from=None):
        """
        Load the language list for TTS.

        text and language parameters are only needed with listen parameter.
        """
        try:
            self.voice_loading = True
            self.tts = TTS[Settings.get().active_tts]()
            self.tts_langs = self.tts.languages
            if not listen:
                GLib.idle_add(self.toggle_voice_spinner, False)
            elif language in self.tts_langs and text != '':
                self.voice_download(text, language)

        except RuntimeError as exc:
            GLib.idle_add(self.on_listen_failed, called_from)
            logging.error('Error: ' + str(exc))
        finally:
            if not listen:
                self.voice_loading = False

    def setup_headerbar(self):
        # Left lang selector
        self.src_lang_selector = DialectLangSelector()
        self.src_lang_selector.connect('notify::selected',
                                       self.on_src_lang_changed)
        self.src_lang_selector.connect('user-selection-changed',
                                       self.translation)
        # Set popover selector to button
        self.src_lang_btn.set_popover(self.src_lang_selector)

        # Right lang selector
        self.dest_lang_selector = DialectLangSelector()
        self.dest_lang_selector.connect('notify::selected',
                                        self.on_dest_lang_changed)
        self.dest_lang_selector.connect('user-selection-changed',
                                        self.translation)
        # Set popover selector to button
        self.dest_lang_btn.set_popover(self.dest_lang_selector)

        self.langs_button_box.set_homogeneous(False)

    def setup_translation(self):
        # Left buffer
        self.src_buffer = self.src_text.get_buffer()
        self.src_buffer.set_text(self.launch_text)
        self.src_buffer.connect('changed', self.on_src_text_changed)
        self.src_buffer.connect('end-user-action', self.user_action_ended)
        # Detect typing
        self.src_key_ctrlr.connect('key-pressed', self.update_trans_button)
        self.win_key_ctrlr.connect('key-pressed', self.on_key_event)
        # "Did you mean" links
        self.mistakes_label.connect('activate-link', self.on_mistakes_clicked)

        # Right buffer
        self.dest_buffer = self.dest_text.get_buffer()
        self.dest_buffer.set_text('')
        self.dest_buffer.connect('changed', self.on_dest_text_changed)
        # Translation progress spinner
        self.trans_spinner.hide()
        self.trans_warning.hide()

        # Voice buttons prep-work
        self.src_voice_warning = Gtk.Image.new_from_icon_name('dialog-warning-symbolic')
        self.src_voice_image = Gtk.Image.new_from_icon_name('audio-speakers-symbolic')
        self.src_voice_spinner = Gtk.Spinner()  # For use while audio is running or still loading.

        self.dest_voice_warning = Gtk.Image.new_from_icon_name('dialog-warning-symbolic')
        self.dest_voice_image = Gtk.Image.new_from_icon_name('audio-speakers-symbolic')
        self.dest_voice_spinner = Gtk.Spinner()

        self.toggle_voice_spinner(True)

        self.src_voice_btn.set_visible(Settings.get().active_tts != '')
        self.dest_voice_btn.set_visible(Settings.get().active_tts != '')

    def toggle_mobile_mode(self):
        if self.mobile_mode:
            # Show actionbar
            self.actionbar.set_revealed(True)
            # Change headerbar title
            self.title_stack.set_visible_child_name('label')
            # Change translation box orientation
            self.translator_box.set_orientation(Gtk.Orientation.VERTICAL)
            # Change lang selectors position
            self.src_lang_btn.set_popover(None)
            self.src_lang_btn2.set_popover(self.src_lang_selector)
            self.dest_lang_btn.set_popover(None)
            self.dest_lang_btn2.set_popover(self.dest_lang_selector)
        else:
            # Hide actionbar
            self.actionbar.set_revealed(False)
            # Reset headerbar title
            self.title_stack.set_visible_child_name('selector')
            # Reset translation box orientation
            self.translator_box.set_orientation(Gtk.Orientation.HORIZONTAL)
            # Reset lang selectors position
            self.src_lang_btn2.set_popover(None)
            self.src_lang_btn.set_popover(self.src_lang_selector)
            self.dest_lang_btn2.set_popover(None)
            self.dest_lang_btn.set_popover(self.dest_lang_selector)

    def translate(self, text, src_lang, dest_lang):
        """
        Translates the given text from auto detected language to last used
        language
        """
        # Set src lang to Auto
        if src_lang is None:
            self.src_lang_selector.set_selected('auto', notify=False)
        else:
            self.src_lang_selector.set_selected(src_lang, notify=False)
        if dest_lang is not None and dest_lang in self.translator.languages:
            self.dest_lang_selector.set_selected(dest_lang)
        # Set text to src buffer
        self.src_buffer.set_text(text)
        # Run translation
        self.translation()

    def save_settings(self, *args, **kwargs):
        if not self.is_maximized():
            size = self.get_default_size()
            Settings.get().window_size = (size.width, size.height)
        if self.translator is not None:
            Settings.get().src_langs = self.src_langs
            Settings.get().dest_langs = self.dest_langs
            Settings.get().save_translator_settings()

    def send_notification(self, text, queue=False, action=None, timeout=5, priority=Adw.ToastPriority.NORMAL):
        """
        Display an in-app notification.

        Args:
            text (str): The text or message of the notification.
            queue (bool, optional): If True, the notification will be queued.
            action (dict, optional): A dict containing the action to be called.
        """
        if not queue and self.toast is not None:
            self.toast.dismiss()
        self.toast = Adw.Toast.new(text)
        self.toast.connect('dismissed', self._toast_dismissed)
        if action is not None:
            self.toast.set_button_label(action['label'])
            self.toast.set_action_name(action['name'])
        self.toast.set_timeout(timeout)
        self.toast.set_priority(priority)
        self.toast_overlay.add_toast(self.toast)

    def _toast_dismissed(self, toast):
        self.toast = None

    def toggle_voice_spinner(self, active=True):
        if active:
            self.lookup_action('listen-src').set_enabled(False)
            self.src_voice_btn.set_child(self.src_voice_spinner)
            self.src_voice_spinner.start()

            self.lookup_action('listen-dest').set_enabled(False)
            self.dest_voice_btn.set_child(self.dest_voice_spinner)
            self.dest_voice_spinner.start()
        else:
            src_text = self.src_buffer.get_text(
                self.src_buffer.get_start_iter(),
                self.src_buffer.get_end_iter(),
                True
            )
            self.lookup_action('listen-src').set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and src_text != ''
            )
            self.src_voice_btn.set_child(self.src_voice_image)
            self.src_voice_spinner.stop()

            dest_text = self.dest_buffer.get_text(
                self.dest_buffer.get_start_iter(),
                self.dest_buffer.get_end_iter(),
                True
            )
            self.lookup_action('listen-dest').set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and dest_text != ''
            )
            self.dest_voice_btn.set_child(self.dest_voice_image)
            self.dest_voice_spinner.stop()

    def on_src_lang_changed(self, _obj, _param):
        code = self.src_lang_selector.get_property('selected')
        dest_code = self.dest_lang_selector.get_property('selected')
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )

        if code == dest_code:
            code = self.dest_langs[1] if code == self.src_langs[0] else dest_code
            self.dest_lang_selector.set_selected(self.src_langs[0], notify=False)

        # Disable or enable listen function.
        if self.tts_langs and Settings.get().active_tts != '':
            self.lookup_action('listen-src').set_enabled(
                code in self.tts_langs and src_text != ''
            )

        # Disable or enable switch function.
        self.lookup_action('switch').set_enabled(
            code != 'auto'
        )

        if code in self.translator.languages:
            self.src_lang_label.set_label(get_lang_name(code))
            # Update saved src langs list
            if code in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code)
            elif code.lower() in self.src_langs:
                # Bring lang to the top
                self.src_langs.remove(code.lower())
            elif len(self.src_langs) == 4:
                self.src_langs.pop()
            self.src_langs.insert(0, code)
        elif code == 'auto' and self.translator.supported_features['detection']:
            self.src_lang_label.set_label(_('Auto'))
        else:
            self.src_lang_label.set_label(get_lang_name(code))

        # Rewrite recent langs
        self.src_lang_selector.clear_recent()
        if self.translator.supported_features['detection']:
            self.src_lang_selector.insert_recent('auto', _('Auto'))
        for code in self.src_langs:
            if code in self.translator.languages:
                self.src_lang_selector.insert_recent(code, get_lang_name(code))

        # Refresh list
        self.src_lang_selector.refresh_selected()

    def on_dest_lang_changed(self, _obj, _param):
        code = self.dest_lang_selector.get_property('selected')
        src_code = self.src_lang_selector.get_property('selected')
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )

        if code == src_code:
            self.src_lang_selector.set_selected(self.dest_langs[0], notify=False)

        # Disable or enable listen function.
        if self.tts_langs and Settings.get().active_tts != '':
            self.lookup_action('listen-dest').set_enabled(
                code in self.tts_langs and dest_text != ''
            )

        self.dest_lang_label.set_label(get_lang_name(code))
        # Update saved dest langs list
        if code in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code)
        elif code.lower() in self.dest_langs:
            # Bring lang to the top
            self.dest_langs.remove(code.lower())
        elif len(self.src_langs) == 4:
            self.dest_langs.pop()
        self.dest_langs.insert(0, code)

        # Rewrite recent langs
        self.dest_lang_selector.clear_recent()
        for code in self.dest_langs:
            if code in self.translator.languages:
                self.dest_lang_selector.insert_recent(code, get_lang_name(code))

        # Refresh list
        self.dest_lang_selector.refresh_selected()

    """
    User interface functions
    """

    def ui_return(self, _action, _param):
        """Go back one step in history."""
        if self.current_history != TRANS_NUMBER:
            self.current_history += 1
            self.history_update()

    def ui_forward(self, _action, _param):
        """Go forward one step in history."""
        if self.current_history != 0:
            self.current_history -= 1
            self.history_update()

    def add_history_entry(self, src_text, src_language, dest_language, dest_text):
        """Add a history entry to the history list."""
        new_history_trans = {
            'Languages': [src_language, dest_language],
            'Text': [src_text, dest_text]
        }
        if self.current_history > 0:
            del self.translator.history[: self.current_history]
            self.current_history = 0
        if len(self.translator.history) == TRANS_NUMBER:
            self.translator.history.pop()
        self.translator.history.insert(0, new_history_trans)
        GLib.idle_add(self.reset_return_forward_btns)

    def switch_all(self, src_language, dest_language, src_text, dest_text):
        self.src_lang_selector.set_selected(dest_language, notify=False)
        self.dest_lang_selector.set_selected(src_language, notify=False)
        self.src_buffer.set_text(dest_text)
        self.dest_buffer.set_text(src_text)
        self.add_history_entry(src_language, dest_language, src_text, dest_text)

        # Re-enable widgets
        self.langs_button_box.set_sensitive(True)
        self.lookup_action('translation').set_enabled(self.src_buffer.get_char_count() != 0)

    def ui_switch(self, _action, _param):
        # Get variables
        self.langs_button_box.set_sensitive(False)
        self.lookup_action('translation').set_enabled(False)
        src_language = self.src_lang_selector.get_property('selected')
        dest_language = self.dest_lang_selector.get_property('selected')
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        if src_language == 'auto':
            return

        # Switch all
        self.switch_all(src_language, dest_language, src_text, dest_text)

    def ui_clear(self, _action, _param):
        self.src_buffer.set_text('')
        self.src_buffer.emit('end-user-action')

    def ui_copy(self, _action, _param):
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        clipboard = Gdk.Display.get_default().get_clipboard()
        clipboard.set(dest_text)

    def ui_paste(self, _action, _param):
        clipboard = Gdk.Display.get_default().get_clipboard()

        def on_paste(_clipboard, result):
            text = clipboard.read_text_finish(result)
            if text is not None:
                end_iter = self.src_buffer.get_end_iter()
                self.src_buffer.insert(end_iter, text)

        cancellable = Gio.Cancellable()
        clipboard.read_text_async(cancellable, on_paste)

    def ui_suggest(self, _action, _param):
        self.dest_toolbar_stack.set_visible_child_name('edit')
        self.before_suggest = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        self.dest_text.set_editable(True)

    def ui_suggest_ok(self, _action, _param):
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )

        (data, headers) = self.translator.format_suggestion(
            self.translator.history[self.current_history]['Text'][0],
            self.translator.history[self.current_history]['Languages'][0],
            self.translator.history[self.current_history]['Languages'][1],
            dest_text
        )
        message = Session.create_message(
            'POST',
            self.translator.suggest_url,
            data, headers
        )
        Session.get().send_and_read_async(message, 0, None, self.on_suggest_response)

        self.before_suggest = None

    def ui_suggest_cancel(self, _action, _param):
        self.dest_toolbar_stack.set_visible_child_name('default')
        if self.before_suggest is not None:
            self.dest_buffer.set_text(self.before_suggest)
            self.before_suggest = None
        self.dest_text.set_editable(False)

    def on_suggest_response(self, session, result):
        success = False
        self.dest_toolbar_stack.set_visible_child_name('default')
        try:
            data = Session.get_response(session, result)
            success = self.translator.get_suggestion(data)
        except Exception as exc:
            logging.error(exc)

        if success:
            self.send_notification(_('New translation has been suggested!'))
        else:
            self.send_notification(_('Suggestion failed.'))

        self.dest_text.set_editable(False)

    def ui_src_voice(self, _action, _param):
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        src_language = self.src_lang_selector.get_property('selected')
        self._voice(src_text, src_language, 'src')

    def ui_dest_voice(self, _action, _param):
        dest_text = self.dest_buffer.get_text(
            self.dest_buffer.get_start_iter(),
            self.dest_buffer.get_end_iter(),
            True
        )
        dest_language = self.dest_lang_selector.get_property('selected')
        self._voice(dest_text, dest_language, 'dest')

    def _voice(self, text, lang, called_from):
        if text != '':
            self.toggle_voice_spinner(True)
            if self.tts_langs:
                threading.Thread(
                    target=self.voice_download,
                    args=(text, lang, called_from),
                    daemon=True
                ).start()
            else:
                threading.Thread(
                    target=self.load_lang_speech,
                    args=(True, text, lang, called_from),
                    daemon=True
                ).start()

    def on_gst_message(self, _bus, message):
        if message.type == Gst.MessageType.EOS:
            self.player.set_state(Gst.State.NULL)
            self.player_event.set()
        elif message.type == Gst.MessageType.ERROR:
            self.player.set_state(Gst.State.NULL)
            self.player_event.set()
            logging.error('Some error occurred while trying to play.')

    def voice_download(self, text, language, called_from):
        try:
            self.voice_loading = True

            with NamedTemporaryFile() as file_to_play:
                self.tts.download_voice(text, language, file_to_play)
                self.player.set_property('uri', 'file://' + file_to_play.name)
                self.player.set_state(Gst.State.PLAYING)
                self.player_event.wait()
        except Exception as exc:
            logging.error(exc)
            logging.error('Audio download failed.')
            GLib.idle_add(self.on_listen_failed, called_from)
        else:
            GLib.idle_add(self.toggle_voice_spinner, False)
        finally:
            self.voice_loading = False

    def on_key_event(self, _button, keyval, _keycode, state):
        modifiers = state & Gtk.accelerator_get_default_mod_mask()
        shift_mask = Gdk.ModifierType.SHIFT_MASK
        unicode_key_val = Gdk.keyval_to_unicode(keyval)
        if (GLib.unichar_isgraph(chr(unicode_key_val))
                and modifiers in (shift_mask, 0)
                and not self.dest_text.get_editable()
                and not self.src_text.is_focus()):
            self.src_text.grab_focus()
            end_iter = self.src_buffer.get_end_iter()
            self.src_buffer.insert(end_iter, chr(unicode_key_val))
            return Gdk.EVENT_STOP
        return Gdk.EVENT_PROPAGATE

    # This starts the translation if Ctrl+Enter button is pressed
    def update_trans_button(self, _button, keyval, _keycode, state):
        modifiers = state & Gtk.accelerator_get_default_mod_mask()

        control_mask = Gdk.ModifierType.CONTROL_MASK
        enter_keys = (Gdk.KEY_Return, Gdk.KEY_KP_Enter)

        if not Settings.get().live_translation:
            if control_mask == modifiers:
                if keyval in enter_keys:
                    if not Settings.get().translate_accel_value:
                        self.translation()
                        return Gdk.EVENT_STOP
                    return Gdk.EVENT_PROPAGATE
            elif keyval in enter_keys:
                if Settings.get().translate_accel_value:
                    self.translation()
                    return Gdk.EVENT_STOP
                return Gdk.EVENT_PROPAGATE

        return Gdk.EVENT_PROPAGATE

    def on_mistakes_clicked(self, _button, _data):
        self.mistakes.set_reveal_child(False)
        self.src_buffer.set_text(self.trans_mistakes[1])
        # Run translation again
        self.translation()

        return Gdk.EVENT_STOP

    def on_src_text_changed(self, buffer):
        sensitive = buffer.get_char_count() != 0
        self.lookup_action('translation').set_enabled(sensitive)
        self.lookup_action('clear').set_enabled(sensitive)
        if not self.voice_loading and self.tts_langs:
            self.lookup_action('listen-src').set_enabled(
                self.src_lang_selector.get_property('selected') in self.tts_langs
                and sensitive
            )
        elif not self.voice_loading and not self.tts_langs:
            self.lookup_action('listen-src').set_enabled(sensitive)

    def on_dest_text_changed(self, buffer):
        sensitive = buffer.get_char_count() != 0
        self.lookup_action('copy').set_enabled(sensitive)
        self.lookup_action('suggest').set_enabled(
            self.translator.supported_features['suggestions']
            and sensitive
        )
        if not self.voice_loading and self.tts_langs:
            self.lookup_action('listen-dest').set_enabled(
                self.dest_lang_selector.get_property('selected') in self.tts_langs
                and sensitive
            )
        elif not self.voice_loading and not self.tts_langs:
            self.lookup_action('listen-dest').set_enabled(sensitive)

    def user_action_ended(self, buffer):
        # If the text is over the highest number of characters allowed, it is truncated.
        # This is done for avoiding exceeding the limit imposed by translation services.
        if buffer.get_char_count() >= self.translator.chars_limit:
            self.send_notification(_('{} characters limit reached!').format(self.translator.chars_limit))
            buffer.delete(
                buffer.get_iter_at_offset(self.translator.chars_limit),
                buffer.get_end_iter()
            )
        self.char_counter.set_text(f'{str(buffer.get_char_count())}/{self.translator.chars_limit}')
        if Settings.get().live_translation:
            self.translation()

    # The history part
    def reset_return_forward_btns(self):
        self.lookup_action('back').set_enabled(self.current_history < len(self.translator.history) - 1)
        self.lookup_action('forward').set_enabled(self.current_history > 0)

    # Retrieve translation history
    def history_update(self):
        self.reset_return_forward_btns()
        lang_hist = self.translator.history[self.current_history]
        self.src_lang_selector.set_selected(lang_hist['Languages'][0], notify=False)
        self.dest_lang_selector.set_selected(lang_hist['Languages'][1], notify=False)
        self.src_buffer.set_text(lang_hist['Text'][0])
        self.dest_buffer.set_text(lang_hist['Text'][1])

    def appeared_before(self):
        src_language = self.src_lang_selector.get_property('selected')
        dest_language = self.dest_lang_selector.get_property('selected')
        src_text = self.src_buffer.get_text(
            self.src_buffer.get_start_iter(),
            self.src_buffer.get_end_iter(),
            True
        )
        if (
            len(self.translator.history) >= self.current_history + 1
            and (self.translator.history[self.current_history]['Languages'][0] == src_language or 'auto')
            and self.translator.history[self.current_history]['Languages'][1] == dest_language
            and self.translator.history[self.current_history]['Text'][0] == src_text
        ):
            return True
        return False

    def translation(self, _action=None, _param=None):
        # If it's like the last translation then it's useless to continue
        if not self.appeared_before():
            src_text = self.src_buffer.get_text(
                self.src_buffer.get_start_iter(),
                self.src_buffer.get_end_iter(),
                True
            )
            src_language = self.src_lang_selector.get_property('selected')
            dest_language = self.dest_lang_selector.get_property('selected')

            if self.ongoing_trans:
                self.next_trans = {
                    'text': src_text,
                    'src': src_language,
                    'dest': dest_language
                }
                return

            if self.next_trans:
                src_text = self.next_trans['text']
                src_language = self.next_trans['src']
                dest_language = self.next_trans['dest']
                self.next_trans = {}

            # Show feedback for start of translation.
            self.translation_loading()

            # If the two languages are the same, nothing is done
            if src_language != dest_language:
                if src_text != '':
                    self.ongoing_trans = True

                    # Format data
                    url = self.translator.translate_url.format(
                        text=src_text, src=src_language, dest=dest_language
                    )
                    (method, data, headers, raw) = self.translator.format_translation(
                        src_text, src_language, dest_language
                    )

                    message = Session.create_message(method, url, data, headers, raw)

                    Session.get().send_and_read_async(
                        message,
                        0,
                        None,
                        self.on_translation_response,
                        (src_text, src_language, dest_language, raw)
                    )
                else:
                    self.trans_failed = False
                    self.trans_mistakes = [None, None]
                    self.trans_src_pron = None
                    self.trans_dest_pron = None
                    self.dest_buffer.set_text('')

                    if not self.ongoing_trans:
                        self.translation_done()

    def on_translation_response(self, session, result, original):
        error = ''
        dest_text = ''

        self.trans_mistakes = [None, None]
        self.trans_src_pron = None
        self.trans_dest_pron = None
        try:
            data = Session.get_response(session, result, raw=original[3])
            (translation, lang) = self.translator.get_translation(data)

            if lang and self.src_lang_selector.get_property('selected') == 'auto':
                if Settings.get().live_translation:
                    lang = get_lang_name(lang)
                    label = _('Auto')
                    self.src_lang_label.set_label(f'{label} ({lang})')
                else:
                    self.src_lang_selector.set_selected(lang, notify=False)

            dest_text = translation.text
            self.trans_mistakes = translation.extra_data['possible-mistakes']
            self.trans_src_pron = translation.extra_data['src-pronunciation']
            self.trans_dest_pron = translation.extra_data['dest-pronunciation']
            self.trans_failed = False

            self.dest_buffer.set_text(dest_text)

            # Finally, everything is saved in history
            self.add_history_entry(
                original[0],
                original[1],
                original[2],
                dest_text
            )
        except ResponseError as exc:
            logging.error(exc)
            error = 'network'
            self.trans_failed = True
        except InvalidApiKey as exc:
            logging.error(exc)
            error = 'invalid-api'
            self.trans_failed = True
        except ApiKeyRequired as exc:
            logging.error(exc)
            error = 'api-required'
            self.trans_failed = True
        except Exception as exc:
            logging.error(exc)
            self.trans_failed = True

        # Mistakes
        if self.translator.supported_features['mistakes'] and not self.trans_mistakes == [None, None]:
            mistake_text = self.trans_mistakes[0].replace('<em>', '<b>').replace('</em>', '</b>')
            self.mistakes_label.set_markup(_('Did you mean: ') + f'<a href="#">{mistake_text}</a>')
            self.mistakes.set_reveal_child(True)
        elif self.mistakes.get_reveal_child():
            self.mistakes.set_reveal_child(False)

        # Pronunciation
        reveal = Settings.get().show_pronunciation
        if self.translator.supported_features['pronunciation']:
            if self.trans_src_pron is not None and self.trans_mistakes == [None, None]:
                self.src_pron_label.set_text(self.trans_src_pron)
                self.src_pron_revealer.set_reveal_child(reveal)
            elif self.src_pron_revealer.get_reveal_child():
                self.src_pron_revealer.set_reveal_child(False)

            if self.trans_dest_pron is not None:
                self.dest_pron_label.set_text(self.trans_dest_pron)
                self.dest_pron_revealer.set_reveal_child(reveal)
            elif self.dest_pron_revealer.get_reveal_child():
                self.dest_pron_revealer.set_reveal_child(False)

        self.translation_failed(self.trans_failed, error)

        self.ongoing_trans = False
        if self.next_trans:
            self.translation()
        else:
            self.translation_done()

    def translation_loading(self):
        self.trans_spinner.show()
        self.trans_spinner.start()
        self.dest_box.set_sensitive(False)
        self.langs_button_box.set_sensitive(False)

    def translation_done(self):
        self.trans_spinner.stop()
        self.trans_spinner.hide()
        self.dest_box.set_sensitive(True)
        self.langs_button_box.set_sensitive(True)

    def translation_failed(self, failed, error=''):
        if failed:
            self.trans_warning.show()
            self.lookup_action('copy').set_enabled(False)
            self.lookup_action('listen-src').set_enabled(False)
            self.lookup_action('listen-dest').set_enabled(False)

            if error == 'network':
                self.send_notification(
                    _('Translation failed, check for network issues'),
                    action={
                        'label': _('Retry'),
                        'name': 'win.translation',
                    }
                )
            elif error == 'invalid-api':
                self.send_notification(
                    _('The provided API key is invalid'),
                    action={
                        'label': _('Preferences'),
                        'name': 'app.preferences',
                    }
                )
            elif error == 'api-required':
                self.send_notification(
                    _('API key is required to use the service'),
                    action={
                        'label': _('Preferences'),
                        'name': 'app.preferences',
                    }
                )
            else:
                self.send_notification(
                    _('Translation failed'),
                    action={
                        'label': _('Retry'),
                        'name': 'win.translation',
                    }
                )
        else:
            self.trans_warning.hide()

    def reload_backends(self):
        self.set_property('backend-loading', True)

        # Load translator
        self.load_translator()
